/*
 * Copyright (c) 2013 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Jan 16 2013
 */
#include "acrbulkit.h"

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/select.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <libgen.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>


/* --- gobal data ---------------------------------------------------------- */

char bind_addr[MAXPATHLEN] = "/var/run";
int disconnect = 0;


/* --- usage --------------------------------------------------------------- */

/** Print usage on a channel.
 *
 * \param[in] channel   Output channel
 * \param[in] argv0     Program invocation name
 */

static void
usage(FILE *channel, char *argv0)
{
  fprintf(channel,
          "usage:  %s [-h] [-u] [-b path] job [job...]\n", basename(argv0));
}


/* --- setup_socket --------------------------------------------------------- */

static int
setup_socket(const char *bind_addr)
{
  union {
    struct sockaddr_un un;
    struct sockaddr addr;
  } addr;
  int s;

  s = socket(PF_LOCAL, SOCK_STREAM, 0);
  if (s < 0) err(2, "socket");

  addr.un.sun_family = PF_LOCAL;
  strncpy(addr.un.sun_path, bind_addr,
          sizeof(addr.un.sun_path)-sizeof("/rbulk-dispatchd"));
  strcat(addr.un.sun_path, "/rbulk-dispatchd");

  if (connect(s, &addr.addr, sizeof(addr.un)) < 0)
    err(2, "connect on %s", addr.un.sun_path);

  return s;
}



/* --- main ---------------------------------------------------------------- */

int
main(int argc, char *const argv[])
{
  extern char *optarg;
  extern int optind;
  char *argv0 = argv[0];

  struct iovec iov[2];
  struct msghdr msg;
  char buf[4096];
  ssize_t l, e;
  int s, c;

  /* parse command line options */
  while ((c = getopt(argc, argv, "hub:")) != -1) {
    switch (c) {
      case 'b':
        if (!realpath(optarg, bind_addr)) err(2, "%s", optarg);
        break;

      case 'u':
        disconnect = 1;
        break;

      case 'h':
        usage(stdout, argv0);
        exit(0);

      case '?':
      default:
        usage(stderr, argv0);
        exit(2);
    }
  }
  argc -= optind;
  argv += optind;
  if (argc <= 0) {
    usage(stderr, argv0);
    exit(2);
  }

  /* setup control socket */
  s = setup_socket(bind_addr);

  /* send jobs */
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = iov;
  msg.msg_iovlen = 2;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  msg.msg_flags = 0;
  iov[1].iov_base = "\n";
  iov[1].iov_len = sizeof("\n")-1;

  while(*argv) {
    iov[0].iov_base = *argv;
    l = iov[0].iov_len = strlen(*argv);
    l += iov[1].iov_len;

    do {
      e = sendmsg(s, &msg, 0);
    } while(e < 0 && (errno == EINTR || errno == ENOBUFS));
    if (e != l) err(2, "sendmsg");

    argv++;
  }

  /* read output */
  while (!disconnect) {
    do {
      l = recv(s, buf, sizeof buf, 0);
    } while(l < 0 && errno == EINTR);
    if (l < 0) {
      if (errno == EAGAIN || errno == EWOULDBLOCK) continue;
      err(2, "recv");
    }
    if (l == 0) break;

    if (fwrite(buf, l, 1, stdout) != 1)
        ;
  }

  /* cleanup */
  close(s);
  return 0;
}
